"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _provideIIFEStatement = require("./provideIIFEStatement");

var _provideIIFEStatement2 = _interopRequireDefault(_provideIIFEStatement);

var _provideInheritedStatement = require("./provideInheritedStatement");

var _provideInheritedStatement2 = _interopRequireDefault(_provideInheritedStatement);

var _provideCommonDepsStatement = require("./provideCommonDepsStatement");

var _provideCommonDepsStatement2 = _interopRequireDefault(_provideCommonDepsStatement);

var _provideLocalExportsStatement = require("./provideLocalExportsStatement");

var _provideLocalExportsStatement2 = _interopRequireDefault(_provideLocalExportsStatement);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * This function builds YM module definition statement using the tree and returns 
 * AST tree of it. Function receives tree and part of the content which will be 
 * wrapped as arguments and returns the new AST statement. Tree has next format:
 * @example <caption>Tree example 1</caption>
 *  {
 *      exported: "export-name",
 *      common: {
 *          deps: [
 *              "imported-module-name-1",
 *              "imported-module-name-3",
 *              "inherited-module-1"
 *          ],
 *          reqs: [
 *              "importedModuleName1",
 *              "importedModuleName3",
 *              "inheritedModule1"
 *          ]
 *      },
 *      bem: {
 *          deps: [
 *              "inherited-module-2",
 *              "inherited-module-3",
 *              "imported__module-2",
 *              "imported__module_for_reexport-1",
 *          ],
 *          reqs: [
 *              "_provide",
 *              "inheritedModule2",
 *              "inheritedModule3",
 *              "importedModuleName2",
 *              "importedModuleForReexport1",
 *              "_self"
 *          ]
 *      },
 *      inheritances: [
 *          "inheritedModule1",
 *          "inheritedModule2",
 *          "inheritedModule3",
 *          "_self"
 *      ],
 *      imports: {
 *          importAlias1: {
 *              member: "importedModuleName1",
 *              property: "property1"
 *          },
 *          importAlias2: {
 *              member: "importedModuleName1",
 *              property: "property2"
 *          },
 *          importAlias3: {
 *              member: "importedModuleName2",
 *              property: null
 *          },
 *          importAlias4: {
 *              member: "importedModuleName3",
 *              property: "default"
 *          },
 *          reexportName1: {
 *              member: "importedModuleForReexport1",
 *              property: "reexportedPropertyName"
 *              
 *          }
 *      },
 *      exports: {
 *          default: "localName1",
 *          exportAlias1: "localName2",
 *          reexportAlias3: "reexportName1"
 *      }
 *  }
 * 
 * The "exported" field contains export name of current module. Using this name we 
 * will be able to import current module module in other modules. For example, if 
 * this field has value "current__module" than in other module we should to specify 
 * this name in import directive to obtain this module in the other module.
 * 
 * @example <caption>Example of obtaining current module with in other</caption>
 *   import currentModule from "bem:current__module"
 * 
 * The "common" field contains the descriptor which describes real imported modules names and
 * names of constants which will generated in the current module to obtain imported modules
 * using CommonJS modular system. For example if the descriptor is
 * @example <caption>Example of descriptor</caption>
 *   {
 *      deps: ["dependency-1"],
 *      reqs: ["dependency1"]
 *   }
 * 
 * 
 * then the code will generated
 * @example <caption>Example of generated code for common descriptor</caption>
 *  const dependency1 = require("dependency-1")
 * 
 * The count of elements in the "deps" field must be equal to count of elements in the "reqs"
 * field.
 * 
 * The "bem" field contains the descriptor which describes real imported modules names and
 * names of constants which will generated in the current module to obtain imported modules
 * using YM modular system and BEM methodology. For example if the descriptor is the 
 * @example <caption>Example of descriptor</caption>
 *   {
 *      deps: ["dependency-1"],
 *      reqs: [
 *          "_provide",
 *          "dependency1",
 *          "_self"
 *      ]
 *   }
 * 
 * then the code will generated
 * @example <caption>Example of generated code for bem descriptor</caption>
 *  modules.define("<exported-name>", ["dependency-1"], (_provide, dependency1, _self) {...;_provide(...)})
 * 
 * The count of elements in the "deps" field must be 2 less then count of elements in the "reqs"
 * field. It is required that first element in the "reqs" field must have "_provide" value and
 * last element must have "_self" value.
 * 
 * The "inheritance" field enumerates names of modules which this module inherits. It is 
 * required that enumerated names match with values in the "reqs" sections and last element
 * was been "_self", but "_provide" element must not be enumerated. For example if we have
 * an descriptor
 * @example <caption>Example of inheritance descriptor</caption>
 *  ["module1", "module2", "_self"]
 * 
 * then the code will generated
 * @example <caption>Example of generated code for previous descriptor</caption>
 *  Object.assign(module1, module2, _self, ...)
 * 
 * The "imports" section contains descriptor which describes matches between 
 * import aliases and imported entities. For example if we have the descriptor
 * @example <caption>Example of imports descriptor</caption>
 *  {
 *      alias: {
 *          member: "moduleName",
 *          property: "property"
 *      }
 *  }
 * 
 * then the code will generated
 * @example <caption>Example of generated code for imports descriptor</caption>
 *  function (alias) {...}(moduleName.property)
 * 
 * The "property" field is optional and if this field will be not defined or has null 
 * value then the code will generated
 * @example <caption>Example of generated code for imports descriptor without property</caption>
 *  function (alias) {...}(moduleName)
 * 
 * The export section contains descriptor which describes matches between export
 * aliases and exported entities. For example if we have the descriptor
 * @example <caption>Example of exports descriptor</caption>
 *  {
 *      alias: "exportedLocalEntity"
 *  }
 * 
 * then the code will generated
 * 
 * @example <caption>Example of generated code for imports descriptor</caption>
 *      return {
 *          alias: exportedLocalEntity
 *      }
 * 
 * Finally, tree describes relations between elements of generated code. The tree, 
 * described above will be transformed by this function to the next code
 * 
 * @example <caption>The code which will be generated using tree described above</caption>
 *  modules.define("export-name", [
 *      "inherited-module-2",
 *      "inherited-module-3",
 *      "imported__module-2",
 *      "imported__module_for_reexport-1"
 *  ], (_provide, inheritedModule1, inheritedModule3, importedModuleName2, importedModuleForReexport1) => {
 *      const importedModuleName1 = require("imported-module-name-1"),
 *            importedModuleName3 = require("imported-module-name-3"),
 *            inheritedModule1 = require("inherited-module-1")
 *      _provide(
 *          Object.assign(
 *              inheritedModule1, 
 *              inheritedModule2, 
 *              inheritedModule3, 
 *              _self, 
 *              function (importAlias1, importAlias2, importAlias3, importAlias4, reexportName1) {
 *                  // Wrapped code...
 *                  return {
 *                      default: localName1,
 *                      exportAlias1: localName2,
 *                      exportAlias3: reexportName1
 *                  }
 *              } (
 *                  importedModuleName1.property1, 
 *                  importedModuleName1.property2,
 *                  importedModuleName2,
 *                  importedModuleName3.default,
 *                  importedModuleForReexport1.reexportedPropertyName
 *          )
 *      )  
 *  })
 * 
 * 
 * 
 * @param {Object} t AST babel builder
 * @param {Object} tree object which stores information about imports and exports
 * @param {Array} content is the part of AST tree which will be wrapped by this function
 * @returns {Object} AST tree of module definition statement
 * @throws {ReferenceError} will throws in the one of next situations:
 *      - If argument t is undefined or has null value
 *      - There is not field "exported" in the tree or has null value
 *      - There are not fields "bem" or/and "common" in the tree or have null value
 *      - There are not fields "deps" or/and "reqs" in the "bem" and/or "common" sections
 *      - There is not field "inheritances" in the tree or has null value
 *      - There is not field "imports" in the tree or has null value
 *      - There is not field "exports" in the tree or has null value
 * @throws {RangeError} if count of dependencies is not equal to count of requirement names 
 * for common section or if count of dependencies is not 2 less count of requirement for
 * bem section which are described in the tree
 * @author Dmitry Mikhaylenko
 */
exports.default = function (t, tree, content) {
    if (!t || !tree) {
        throw new ReferenceError("Arguments must not have null or undefined value");
    } else if (!tree.exported || !tree.bem) {
        throw new ReferenceError("Exports name and bem deps sections must be specified in the tree");
    } else if (!tree.bem.deps || !tree.bem.reqs) {
        throw new ReferenceError("Both dependencies and requirement names must be specified in the bem section and can not have null or undefined values");
    } else if (tree.bem.deps.length !== tree.bem.reqs.length - 2) {
        throw new RangeError("The count of dependecies must be 2 less than count of requirement names");
    } else {
        return (0, _provideLocalExportsStatement2.default)(t, tree.exports, function (localExportsStatement) {
            return (0, _provideCommonDepsStatement2.default)(t, tree.common, function (commonDepsStatement) {
                var wrapped = content ? content.concat([localExportsStatement]) : [localExportsStatement];
                return (0, _provideIIFEStatement2.default)(t, tree.imports, wrapped, function (iife) {
                    return (0, _provideInheritedStatement2.default)(t, tree.inheritances, iife, function (inheritedStatement) {
                        var provideStatement = t.expressionStatement(t.callExpression(t.identifier(tree.bem.reqs[0]), [inheritedStatement]));

                        var depsStatement = tree.bem.deps.map(function (dep) {
                            return t.stringLiteral(dep);
                        });

                        var depsParams = tree.bem.reqs.map(function (req) {
                            return t.identifier(req);
                        });

                        var depsArrayStatement = t.arrayExpression(depsStatement);

                        var provideCallbackStatement = t.arrowFunctionExpression(depsParams, t.blockStatement(commonDepsStatement ? [commonDepsStatement, provideStatement] : [provideStatement]));

                        return t.expressionStatement(t.callExpression(t.memberExpression(t.identifier("modules"), t.identifier("define")), [t.stringLiteral(tree.exported), depsArrayStatement, provideCallbackStatement]));
                    });
                });
            });
        });
    }
};

module.exports = exports["default"];